<!doctype html>
<html lang="en">
<head>
    <meta charset="utf-8">
    <title>mitmproxy</title>
    <link rel="icon" href=".{{ static_url('images/favicon.ico') }}" type="image/x-icon"/>
    <meta name="viewport" content="width=device-width, initial-scale=1" />
    <style>
        body {
            font-family: sans-serif;
            display: flex;
            flex-direction: column;
            align-items: center;
        }
        input {
            font-family: monospace;
        }
    </style>
</head>
<body>
    {% if invalid_password %}
    <h1 style="color: darkred">403 Invalid Password</h1>
    {% else %}
    <h1>403 Authentication Required</h1>
    {% end %}
    <p>To access mitmproxy, please enter the password or authentication token printed in the console.</p>
    <form method="POST">
        <label>
            <input type="password" name="token" size="32" placeholder="" />
        </label>
        {% module xsrf_form_html() %}
        <input type="submit" />
    </form>
</body>
</html>
